import torch
from torch import nn
from .loss_interface import AbstractRegressionLoss


class CrossEntropyLoss(AbstractRegressionLoss):
    """
    This class computes the cross entropy loss between two tensors.
    """

    def _initialize_loss_function_object(self):
        return nn.CrossEntropyLoss(reduction=self.reduction_method)


class BinaryCrossEntropyLoss(AbstractRegressionLoss):
    """
    This class computes the binary cross entropy loss between two tensors.
    """

    def _initialize_loss_function_object(self):
        return nn.BCELoss(reduction=self.reduction_method)


class BinaryCrossEntropyWithLogitsLoss(AbstractRegressionLoss):
    """
    This class computes the binary cross entropy loss with logits between two tensors.
    """

    def _initialize_loss_function_object(self):
        return nn.BCEWithLogitsLoss(reduction=self.reduction_method)


class BaseLossWithScaledTarget(AbstractRegressionLoss):
    """
    General interface for the loss functions requiring scaling of the target tensor.
    """

    def _initialize_scaling_factor(self):
        loss_params: dict = self.params["loss_function"]
        self.scaling_factor = loss_params.get("scaling_factor", 1.0)
        if isinstance(loss_params, dict):
            self.scaling_factor = loss_params.get("scaling_factor", self.scaling_factor)
        return self.scaling_factor

    def _calculate_loss(self, prediction: torch.Tensor, target: torch.Tensor):
        return self.loss_calculator(prediction, target * self.scaling_factor)


class L1Loss(BaseLossWithScaledTarget):
    """
    This class computes the L1 loss between two tensors.
    """

    def _initialize_loss_function_object(self):
        return nn.L1Loss(reduction=self.reduction_method)


class MSELoss(BaseLossWithScaledTarget):
    """
    This class computes the mean squared error loss between two tensors.
    """

    def _initialize_loss_function_object(self):
        return nn.MSELoss(reduction=self.reduction_method)



#!usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import click
from deprecated import deprecated
from typing import Optional

from GANDLF import version
from GANDLF.cli import copyrightMessage
from GANDLF.cli.generate_metrics import generate_metrics_dict
from GANDLF.entrypoints import append_copyright_to_help
from GANDLF.utils import logger_setup


def _generate_metrics(
    input_data: str,
    config: str,
    output_file: Optional[str],
    missing_prediction: int = -1,
):
    generate_metrics_dict(input_data, config, output_file, missing_prediction)
    print("Finished.")


@click.command()
@click.option(
    "--config",
    "-c",
    required=True,
    help="The configuration file (contains all the information related to the training/inference session)",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
)
@click.option(
    "--input-data",
    "-i",
    required=True,
    type=str,
    help="The CSV file of input data that is used to generate the metrics; "
    "should contain 3 columns: 'SubjectID,Target,Prediction'",
)
@click.option(
    "--output-file",
    "-o",
    type=click.Path(file_okay=True, dir_okay=False),
    help="Location to save the output dictionary. If not provided, will print to stdout.",
)
@click.option(
    "--missing-prediction",
    "-m",
    required=False,
    type=int,
    default=-1,
    help="The value to use for missing predictions as penalty; if `-1`, this does not get added. This is only used in the case where the targets and predictions are passed independently.",
)
@click.option(
    "--log-file",
    type=click.Path(),
    default=None,
    help="Output file which will contain the logs.",
)
@click.option("--raw-input", hidden=True)
@append_copyright_to_help
def new_way(
    config: str,
    input_data: str,
    output_file: Optional[str],
    missing_prediction: int,
    raw_input: str,
    log_file: str,
):
    """Metrics calculator."""

    logger_setup(log_file)
    _generate_metrics(
        input_data=input_data,
        config=config,
        output_file=output_file,
        missing_prediction=missing_prediction,
    )


@deprecated(
    "This is a deprecated way of running GanDLF. Please, use `gandlf generate-metrics` cli command "
    + "instead of `gandlf_generateMetrics`. Note that in new CLI tool some params were renamed or "
    "changed its behavior:\n"
    + "  --parameters_file to --config\n"
    + "  --inputdata/--data_path to --input-data\n"
    + "  --outputfile/--output_path to --output-file\n"
    + "  --version removed; use `gandlf --version` instead\n"
    + "`gandlf_generateMetrics` script would be deprecated soon."
)
def old_way():
    logger_setup()
    parser = argparse.ArgumentParser(
        prog="GANDLF_Metrics",
        formatter_class=argparse.RawTextHelpFormatter,
        description="Metrics calculator.\n\n" + copyrightMessage,
    )
    parser.add_argument(
        "-c",
        "--config",
        "--parameters_file",
        metavar="",
        type=str,
        required=True,
        help="The configuration file (contains all the information related to the training/inference session)",
    )
    parser.add_argument(
        "-i",
        "--inputdata",
        "--data_path",
        metavar="",
        type=str,
        required=True,
        help="The CSV file of input data that is used to generate the metrics; should contain 3 columns: 'SubjectID,Target,Prediction'",
    )
    parser.add_argument(
        "-o",
        "--outputfile",
        "--output_path",
        metavar="",
        type=str,
        default=None,
        help="Location to save the output dictionary. If not provided, will print to stdout.",
    )
    parser.add_argument(
        "-m",
        "--missingprediction",
        metavar="",
        type=int,
        default=-1,
        help="The value to use for missing predictions as penalty; if `-1`, this does not get added. This is only used in the case where the targets and predictions are passed independently.",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s v{}".format(version) + "\n\n" + copyrightMessage,
        help="Show program's version number and exit.",
    )

    # This is a dummy argument that exists to trigger MLCube mounting requirements.
    # Do not remove.
    parser.add_argument("-rawinput", "--rawinput", help=argparse.SUPPRESS)

    args = parser.parse_args()
    assert args.config is not None, "Missing required parameter: config"
    assert args.inputdata is not None, "Missing required parameter: inputdata"

    _generate_metrics(
        input_data=args.inputdata,
        config=args.config,
        output_file=args.outputfile,
        missing_prediction=args.missingprediction,
    )


if __name__ == "__main__":
    old_way()



import torch
from torch import nn
from abc import ABC, abstractmethod
from typing import List


class AbstractLossFunction(nn.Module, ABC):
    def __init__(self, params: dict):
        nn.Module.__init__(self)
        self.params = params
        self.num_classes = len(params["model"]["class_list"])
        self._initialize_penalty_weights()

    def _initialize_penalty_weights(self):
        default_penalty_weights = torch.ones(self.num_classes)
        self.penalty_weights = self.params.get(
            "penalty_weights", default_penalty_weights
        )

    @abstractmethod
    def forward(self, prediction: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the loss function. To be implemented by child classes.
        """


class AbstractSegmentationLoss(AbstractLossFunction):
    """
    Base class for loss functions that are used for segmentation tasks.
    """

    def __init__(self, params: dict):
        super().__init__(params)

    def _compute_single_class_loss(
        self, prediction: torch.Tensor, target: torch.Tensor, class_idx: int
    ) -> torch.Tensor:
        """Compute loss for a single class."""
        loss_value = self._single_class_loss_calculator(
            prediction[:, class_idx, ...], target[:, class_idx, ...]
        )
        return 1 - loss_value

    def _optional_loss_operations(self, loss: torch.Tensor) -> torch.Tensor:
        """
        Perform additional operations on the loss value. Defaults to identity operation.
        If needed, child classes can override this method. Useful in cases where
        for example, the loss value needs to log-transformed or clipped.
        """
        return loss

    @abstractmethod
    def _single_class_loss_calculator(
        self, prediction: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        """
        Compute loss for a pair of prediction and target tensors. To be implemented by child classes.
        """

    def forward(self, prediction: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        accumulated_loss = torch.tensor(0.0, device=prediction.device)

        for class_idx in range(self.num_classes):
            current_loss = self._compute_single_class_loss(
                prediction, target, class_idx
            )
            accumulated_loss += (
                self._optional_loss_operations(current_loss)
                * self.penalty_weights[class_idx]
            )

        accumulated_loss /= self.num_classes

        return accumulated_loss


class AbstractRegressionLoss(AbstractLossFunction):
    """
    Base class for loss functions that are used for regression and classification tasks.
    """

    def __init__(self, params: dict):
        super().__init__(params)
        self.loss_calculator = self._initialize_loss_function_object()
        self.reduction_method = self._initialize_reduction_method()

    def _initialize_reduction_method(self) -> str:
        """
        Initialize the reduction method for the loss function. Defaults to 'mean'.
        """
        loss_params = self.params["loss_function"]
        reduction_method = "mean"
        if isinstance(loss_params, dict):
            reduction_method = loss_params.get("reduction", reduction_method)
            assert reduction_method in [
                "mean",
                "sum",
            ], f"Invalid reduction method defined for loss function: {reduction_method}. Valid options are ['mean', 'sum']"
        return reduction_method

    def _calculate_loss_for_single_class(
        self, prediction: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        """
        Calculate loss for a single class. To be implemented by child classes.
        """
        return self.loss_calculator(prediction, target)

    @abstractmethod
    def _initialize_loss_function_object(self) -> nn.modules.loss._Loss:
        """
        Initialize the loss function object used in the forward method. Has to return
        callable pytorch loss function object.
        """

    def forward(self, prediction: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        accumulated_loss = torch.tensor(0.0, device=prediction.device)
        for class_idx in range(self.num_classes):
            accumulated_loss += (
                self._calculate_loss_for_single_class(
                    prediction[:, class_idx, ...], target[:, class_idx, ...]
                )
                * self.penalty_weights[class_idx]
            )

        accumulated_loss /= self.num_classes

        return accumulated_loss


class AbstractHybridLoss(AbstractLossFunction):
    """
    Base class for hybrid loss functions that are used for segmentation tasks.
    """

    def __init__(self, params: dict):
        super().__init__(params)
        self.loss_calculators = self._initialize_all_loss_calculators()

    @abstractmethod
    def _initialize_all_loss_calculators(self) -> List[AbstractLossFunction]:
        """
        Each hybrid loss should implement this method, creating all loss functions as a list that
        will be used during the forward pass.
        """
        pass

    def forward(self, prediction: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        accumulated_loss = torch.tensor(0.0, device=prediction.device)
        for loss_calculator in self._initialize_all_loss_calculators():
            accumulated_loss += loss_calculator(prediction, target)

        return accumulated_loss
